@@ -5,3 +5,7 @@ class LanguageNotSupported(Exception):
 class BoundedListEmpty(Exception):
     """Raised when in Concatenator class the BOUNDS are not yet loaded, but concatenate() is tried""" 
     pass
+
+class StanzaPipelineFail(Exception):
+    """Raised when Stanza pipelines fail to load."""
+    pass
